"""
Match making tool.
"""
import abc
import random
from itertools import permutations
from typing import Any

from pydantic import BaseModel

from whist_core.session.distribution import Distribution, DistributionEntry
from whist_core.session.userlist import UserList

subclass_registry = {}


# pylint: disable=too-few-public-methods
class Matcher(abc.ABC, BaseModel, extra='allow'):
    """
    Abstrakt class for player to teams matching.
    """
    teams: list[Distribution] = []
    number_teams: int

    def __init_subclass__(cls, **kwargs: Any) -> None:
        """
        Initializes the subclass
        :param kwargs: field from above
        :return: None
        """
        super().__init_subclass__(**kwargs)
        subclass_registry[cls.__name__] = cls

    @abc.abstractmethod
    def distribute(self, users: UserList) -> Distribution:
        """
        Distributes cards according to subclass implementation.
        :param users: the players to be distributed to teams
        :return: the list of teams with players distributed to them
        """
        raise NotImplementedError

    def _apply_distribution(self, distribution):
        self.teams.append(distribution)


class RoundRobinMatcher(Matcher):
    """
    Distributes the players in the order of the user list.
    """
    iteration: int = 0
    distributions: list[Distribution] = []

    def __init__(self, number_teams: int, **data):
        """
        Constructor. See details in base class.
        :param number_teams:
        :param team_size:
        :param data:
        """
        super().__init__(number_teams=number_teams, **data)

    def distribute(self, users: UserList) -> Distribution:
        """
        Distributes one player to each team each round in order of the user list. Repeats until
        the user list is empty.
        :param users: the players to be distributed to teams
        :return: the teams in round robin distribution
        """
        if len(self.distributions) != len(users):
            self.distributions = []
            self._precalculate_distributions(len(users))
        distribution = self.distributions[self.iteration]
        self.iteration += 1
        self._apply_distribution(distribution)

        return distribution

    def _precalculate_distributions(self, number_players: int):
        for distribution_int in sorted(
                set(permutations((x % self.number_teams for x in range(number_players))))):
            distribution = Distribution()
            for player_index, team_id in enumerate(distribution_int):
                distribution.add(DistributionEntry(player_index=player_index, team_id=team_id))
            self.distributions.append(distribution)


class RandomMatcher(Matcher):
    """
    Distributes the players randomly to teams.
    """

    def distribute(self, users: UserList) -> Distribution:
        """
        For given parameter distributes the players to teams.
        :return: None
        :rtype: None
        """
        players = users.players
        team_size: int = int(len(players) / self.number_teams)
        teams: list = list(range(0, team_size)) * self.number_teams
        distribution: Distribution = Distribution()
        for player_index in range(len(players)):
            team_id = random.choice(teams)  # nosec random
            teams.remove(team_id)
            distribution.add(DistributionEntry(player_index=player_index, team_id=team_id))

        self._apply_distribution(distribution)

        return distribution



"""Trick implementation"""
from pydantic import BaseModel

from whist_core.cards.card import Card, Suit
from whist_core.cards.card_container import OrderedCardContainer
from whist_core.game.errors import NotPlayersTurnError, TrickDoneError, CardNotInHandError
from whist_core.game.legal_checker import LegalChecker
from whist_core.game.player_at_table import PlayerAtTable
from whist_core.game.warnings import TrickNotDoneWarning, ServSuitFirstWarning
from whist_core.util import enforce_str_on_dict


class Trick(BaseModel):
    """
    One round of where every player plays one card.
    """
    play_order: list[PlayerAtTable]
    stack: OrderedCardContainer = OrderedCardContainer.empty()
    trump: Suit

    @property
    def done(self) -> bool:
        """
        Is the trick done.
        :return: True if trick is done else false.
        :rtype: bool
        """
        return len(self.stack) == len(self.play_order)

    @property
    def winner(self) -> PlayerAtTable:
        """
        Player how won the trick.
        :return: Player instance of the winner if the trick is done.
        Else raises TrickNotDoneWarning
        :rtype: Player
        """
        if not self.done:
            raise TrickNotDoneWarning()
        turn, _ = self.stack.get_turn_and_winner_card(self.trump)
        return self.play_order[turn]

    def play_card(self, player: PlayerAtTable, card: Card) -> None:
        """
        One player plays one card. Which is put on top of the stack.
        :param player: Player who wants to play a card.
        :type player: Player
        :param card: Card which the player wants to play.
        :type card: Card
        :return: None if successful, else raises TrickDoneError if every player already played a
        card.
        Or NotPlayersTurnError if a player attempts to play card although it is not they turn.
        :rtype: None
        """
        turn = len(self.stack)
        if turn == len(self.play_order):
            raise TrickDoneError()
        if player != self.play_order[turn]:
            raise NotPlayersTurnError(player.player, self.play_order[turn].player)
        if not LegalChecker.check_legal(player.hand, card, self.stack.first):
            raise ServSuitFirstWarning()
        if card not in player.hand:
            raise CardNotInHandError(f'{card} is not in {player}\'s hand')

        self.stack.add(card)
        player.hand.remove(card)

    def dict(self, *args, **kwargs):
        """Returns dictionary."""
        super_dict = super().dict(*args, **kwargs)
        return enforce_str_on_dict(super_dict, {'trump'})



"""
Collection of utility functions.
"""
from typing import Iterable


def enforce_str_on_dict(dictionary: dict, keys: Iterable[str]) -> dict:
    """
    Forces a dictionary to use string values instead of objects for the given keys.
    :param dictionary: which needs transformation
    :param keys: that values needed to be transformed
    :return: the above dictionary with values changes to string
    """
    for key in keys:
        if key in dictionary:
            dictionary[key] = str(dictionary[key]) if dictionary[key] else None
    return dictionary



"""
Data wrapper of player stati.
"""
from typing import Optional

# Wrapper class
# pylint: disable=too-few-public-methods
from pydantic import BaseModel


class Status(BaseModel):
    """
    Player status a table.
    """
    ready: bool = False
    team: Optional[int] = None



"""Ring buffer of players at the table."""
from typing import Optional

from pydantic import BaseModel

from whist_core.cards.card_container import UnorderedCardContainer
from whist_core.error.table_error import PlayerNotJoinedError
from whist_core.game.player_at_table import PlayerAtTable
from whist_core.scoring.team import Team
from whist_core.user.player import Player


class PlayOrder(BaseModel):
    """
    Iterates over the players at the table.
    """

    next_player: int
    play_order: list[PlayerAtTable]

    def __iter__(self):
        """Iteration over all players."""
        return iter(self.play_order)

    @staticmethod
    def from_team_list(teams: list[Team]):
        """
        Factory method to create a play order from a team list.
        :param teams: list of teams
        :return: PlayOrder
        """
        size = len(teams) * len(teams[0].players)
        play_order: list[Optional[PlayerAtTable]] = [None] * size
        for team_index, team in enumerate(teams):
            for player_index, player in enumerate(team.players):
                player_index = team_index + player_index * len(teams)
                play_order[player_index] = PlayerAtTable(
                    player=player,
                    hand=UnorderedCardContainer.empty(),
                    team=team_index
                )
        return PlayOrder(play_order=play_order, next_player=0)

    def rotate(self, player: PlayerAtTable) -> 'PlayOrder':
        """
        Rotates the play order, so the player will be next player.
        :param player: who should be at beginning of the play order
        :return: None
        """
        order = list(self)
        rotation: int = order.index(player)
        return PlayOrder(play_order=PlayOrder._new_rotate_order(self, rotation), next_player=0)

    def next_order(self) -> 'PlayOrder':
        """
        Create the order for the next hand.
        :rtype: PlayOrder
        """
        return PlayOrder(play_order=PlayOrder._new_order(self), next_player=0)

    def get_next_player(self) -> PlayerAtTable:
        """
        Retrieves the next player who's turn it is.
        :rtype: PlayOrder
        """
        player: PlayerAtTable = self.play_order[self.next_player]
        self.next_player = (self.next_player + 1) % len(self.play_order)
        return player

    def get_player(self, player: Player) -> PlayerAtTable:
        """
        Retrieves the PlayerAtTable for the player given.
        :param player: who needs it's counterpart at the table
        :return: the player at table
        :raises PlayerNoteJoinedError: when a player is requested but is not in play order.
        """
        players_matching = [table_player for table_player in self.play_order if
                            table_player.player == player]
        if len(players_matching) == 0:
            raise PlayerNotJoinedError()
        return players_matching[0]

    def to_team_list(self) -> list[list[Player]]:
        """
        Returns a two-dimensional array of players sorted by teams.
        """
        players_by_team = [[], []]
        for player in self.play_order:
            players_by_team[player.team].append(player.player)
        return players_by_team

    # pylint: disable=protected-access
    @classmethod
    def _new_order(cls, old_order: 'PlayOrder'):
        return old_order.play_order[1:] + old_order.play_order[:1]

    @classmethod
    def _new_rotate_order(cls, old_order: 'PlayOrder', rotation: int):
        return old_order.play_order[rotation:] + old_order.play_order[:rotation]



"""
Elo Rating Calculator
"""
from whist_core.scoring.score_card import ScoreCard
from whist_core.scoring.team import Team
from whist_core.user.player import Player


# pylint: disable=too-few-public-methods
class EloRater:
    """
    Static class that calculates the Elo-Rating for players after several hands played.
    """

    @staticmethod
    def rate(teams: list[Team], scores: ScoreCard) -> None:
        """
        Calculates the new rating of player after several hand played.
        :param teams:
        :type teams:
        :param scores:
        :type scores:
        :return:
        :rtype:
        """
        delta = EloRater._score_delta(teams[0], teams[1], scores)
        for team in teams:
            for player in team.players:
                k_factor = EloRater._k_factor(player)
                won = scores.won(team)
                player.rating += round(k_factor * delta * won)

    @staticmethod
    def _k_factor(player: Player) -> int:
        if player.rating > 2400 and player.games > 30:
            return 10
        if player.rating < 2300 and player.games < 30:
            return 40
        return 20

    @staticmethod
    def _score_delta(team: Team, opponent: Team, scores: ScoreCard) -> float:
        num_games = len(scores)
        num_wins = scores.score(team)
        expected_score = EloRater._expected_score(team, opponent)
        return num_wins - num_games * expected_score

    @staticmethod
    def _expected_score(team: Team, opponent: Team) -> float:
        q_a = EloRater._team_quotient(team)
        q_b = EloRater._team_quotient(opponent)

        return q_a / (q_a + q_b)

    @staticmethod
    def _team_quotient(team: Team):
        return 10 ** (team.rating / 400)



"""Errors regarding table logic."""


class TableSettingsError(Exception):
    """
    Thrown if the table is not setup correctly.
    """


class TableFullError(Exception):
    """
    Thrown if the table is already full.
    """


class TeamFullError(Exception):
    """
    Thrown if the team is already full.
    """


class TableNotReadyError(Exception):
    """
    Is raised when a table requires all player to be ready, but at least one is not.
    """


class TableNotStartedError(Exception):
    """
    Is raised when a table is expected to have started, but it wasn't. Similar to
    TableNotReadyError.
    """


class PlayerNotJoinedError(Exception):
    """
    Raised if a player has not yet joined the table.
    """



"""Extraction of common methods for class that contains a set of cards"""
import abc
import random
from typing import Iterator, Optional

from pydantic import BaseModel, PrivateAttr

from whist_core.cards.card import Card, Suit


class CardContainer(BaseModel, abc.ABC, frozen=True):
    """
    Abstract Base Class for card containers. Duplicate cards are not allowed.
    """

    cards: tuple[Card, ...]

    @classmethod
    def empty(cls) -> 'CardContainer':
        """
        Creates an empty card container.

        :return: empty ard container
        :rtype: correct subtype of CardContainer
        """
        return cls(cards=())

    @classmethod
    def with_cards(cls, *cards) -> 'CardContainer':
        """
        Creates a card container with the given cards.

        :param cards: cards to add
        :return: card container with given cards
        :rtype: correct subtype of CardContainer
        """
        if len(cards) == 1 and not isinstance(cards[0], Card):
            cards = cards[0]
        return cls(cards=cards)

    @classmethod
    def full(cls) -> 'CardContainer':
        """
        Create a full card container.

        :return: full card container
        :rtype: correct subtype of CardContainer
        """
        return cls(cards=Card.all_cards())

    def pop_random(self) -> Card:
        """
        Removes one random card from card container.

        :return: A card from deck.
        """
        card = random.choice(self.cards)  # nosec random
        self.remove(card)
        return card

    def __contains__(self, card: Card) -> bool:
        """Returns if a card is in container. True if yes else False."""
        return card in self.cards

    def __len__(self):
        """Returns the amount of cards in the container."""
        return len(self.cards)

    def __iter__(self) -> Iterator[Card]:
        """Iterates over all cards."""
        return iter(self.cards)

    def __str__(self) -> str:
        """Returns string representation of all cards."""
        return str(self.cards)

    def __repr__(self) -> str:
        """Returns string representation of all cards with class name."""
        return f'{self.__class__.__name__}(cards={self.cards!r})'

    def remove(self, card: Card) -> None:
        """
        Remove a card from this container.

        :param card: card to remove
        """
        if not isinstance(card, Card):
            raise ValueError(f'cannot remove {card} of type {type(card)} from card container')
        if card not in self:
            raise ValueError(f'{card} not in card container')
        self._remove_impl(card)

    @abc.abstractmethod
    def _remove_impl(self, card: Card) -> None:
        raise NotImplementedError

    def add(self, card: Card) -> None:
        """
        Add a card to this container.

        :param card: card to add
        """
        if not isinstance(card, Card):
            raise ValueError(f'cannot add {card} of type {type(card)} to card container')
        if card in self:
            raise ValueError(f'{card} already in card container')
        self._add_impl(card)

    @abc.abstractmethod
    def _add_impl(self, card: Card) -> None:
        raise NotImplementedError

    def contains_suit(self, suit: Suit) -> bool:
        """
        Checks if a card of a suit is still in the card container.

        :param suit: which should be checked
        :return: True if contains this suit else False
        """
        return any(card.suit == suit for card in self)

    def get_cards_of_suit(self, suit: Suit) -> Iterator[Card]:
        """
        Get all cards of a suit in the card container.

        :param suit: which should be checked
        :return: iterator of cards of given suit
        """
        return filter(lambda card: card.suit == suit, self)


class UnorderedCardContainer(CardContainer):
    """
    Base Class unordered card containers
    """

    _cards_set: set[Card] = PrivateAttr()

    def __init__(self, **data):
        """
        Constructor.
        :param data: set of cards
        """
        super().__init__(**data)
        self._cards_set = set(self.cards)
        self.__resync()

    def __contains__(self, card: Card) -> bool:
        """Returns if a card is in container. True if yes else False."""
        return card in self._cards_set

    def _remove_impl(self, card: Card) -> None:
        self._cards_set.remove(card)
        self.__resync()

    def _add_impl(self, card: Card) -> None:
        self._cards_set.add(card)
        self.__resync()

    def __resync(self) -> None:
        """
        de-duplicate and re-sort self.cards - i.e. synchronize with the set representation
        """
        self.model_config['frozen'] = False
        self.cards = tuple(sorted(self._cards_set))
        self.model_config['frozen'] = True


class OrderedCardContainer(CardContainer):
    """
    Base Class ordered card containers
    """

    def _remove_impl(self, card: Card) -> None:
        card_list = list(self.cards)
        card_list.remove(card)

        self.model_config['frozen'] = False
        self.cards = tuple(card_list)
        self.model_config['frozen'] = True

    def _add_impl(self, card: Card) -> None:
        self.model_config['frozen'] = False
        self.cards = (*self.cards, card)
        self.model_config['frozen'] = True

    @property
    def first(self) -> Optional[Card]:
        """
        Returns the first card in the card container.
        :return: The first card played if it exists. Else None.
        """
        if len(self) == 0:
            return None
        return self.cards[0]

    @property
    def last(self) -> Optional[Card]:
        """
        Returns the last card in the card container.
        :return: The last card played if it exists. Else None.
        """
        if len(self) == 0:
            return None
        return self.cards[-1]

    def get_turn(self, card: Card) -> int:
        """
        Gets the turn of a card played.
        :param card: for which the turn number shall be found
        :return: the index of the card in the card container. 0 is the first the card played
        """
        return self.cards.index(card)

    def get_turn_and_winner_card(self, trump: Suit) -> Optional[tuple[int, Card]]:
        """
        Returns the highest trump card or the highest card of the suit played first.
        :param trump: suit of trump
        :return: the winning card and its index
        """
        if len(self) == 0:
            return None
        winner_suit_cards = list(self.get_cards_of_suit(trump))
        if len(winner_suit_cards) == 0:
            winner_suit_cards = self.get_cards_of_suit(self.first.suit)
        winner_card = max(winner_suit_cards, key=lambda x: x.rank)
        return self.get_turn(winner_card), winner_card



"""Rubber of whist"""
from pydantic import BaseModel

from whist_core.game.errors import GameNotStartedError, GameNotDoneError
from whist_core.game.game import Game
from whist_core.game.play_order import PlayOrder
from whist_core.scoring.team import Team


class Rubber(BaseModel, arbitrary_types_allowed=True):
    """
    Implementation of a rubber.
    """
    max_games: int = 3
    games: list[Game] = []
    teams: list[Team]

    @property
    def games_played(self) -> int:
        """
        Amounts of games played already.
        :rtype: int
        """
        return len(self.games)

    @property
    def done(self) -> bool:
        """
        Checks if the rubber is done.
        :return: True if done else False
        :rtype: bool
        """
        return self.games_played == self.max_games

    def current_game(self) -> Game:
        """
        Returns the current game.
        """
        if len(self.games) == 0:
            raise GameNotStartedError()
        return self.games[-1]

    def next_game(self) -> Game:
        """
        Creates a new game if the previous is done.
        :rtype: Game
        """
        if len(self.games) == 0 or self.games[-1].done:
            self.games.append(Game(play_order=PlayOrder.from_team_list(self.teams)))
        elif not self.games[-1].done:
            raise GameNotDoneError()

        return self.current_game()



"""Wraps player's team assignment."""
from pydantic import BaseModel


class DistributionEntry(BaseModel):
    """
    Player to team assignment.
    """
    player_index: int
    team_id: int


class Distribution(BaseModel):
    """
    All player to team assignment.
    """
    entries: list[DistributionEntry] = []

    def __iter__(self):
        """
        Iterates over all entries.
        """
        return iter(self.entries)

    def __len__(self):
        """
        Returns the list of entries.
        """
        return len(self.entries)

    def __getitem__(self, item: int) -> DistributionEntry:
        """
        Gets on specific entry.
        :param item: index of the item
        """
        return self.entries[item]

    def add(self, entry: DistributionEntry) -> None:
        """
        Adds a new entry to the list.
        :param entry: to be added
        :return: None
        """
        self.entries.append(entry)



"""Errors in game phase"""
from whist_core.user.player import Player


class CardNotInHandError(Exception):
    """
    Raised when a player tries to play a card, that is not in their hand.
    """


class HandDoneError(Exception):
    """
    Raised when the next trick is requested, but current Hand is already done..
    """


class HandNotDoneError(Exception):
    """
    Raised when the next hand is requested, but current is not done yet.
    """


class GameNotDoneError(Exception):
    """Raised if next game is requested, but current is not done yet."""


class GameNotStartedError(Exception):
    """
    Raised if a game has not be started.
    """


class NotPlayersTurnError(Exception):
    """
    Raised when players tries to play although is not their turn.
    """

    def __init__(self, player: Player, turn_player: Player):
        """
        Constructor.
        :param player: Who tried to play a card.
        :param turn_player: Which turn it actually is.
        """
        super().__init__()
        self.message = f'Is not {player} turn, but {turn_player}.'


class NoTrumpSelectedError(Exception):
    """
    Raised when no trump is selected for hand.
    """


class RubberNotDoneError(Exception):
    """
    Raised if the current rubber is not done, but action requested requires it to be done.
    """


class TrickDoneError(Exception):
    """
    Raised when the trick is already done.
    """



"""
Handles users joining and leaving a table.
"""
from itertools import groupby
from typing import Optional, Dict

from pydantic import BaseModel

from whist_core.error.table_error import PlayerNotJoinedError
from whist_core.scoring.team import Team
from whist_core.session.distribution import Distribution
from whist_core.user.player import Player
from whist_core.user.status import Status


class UserListEntry(BaseModel):
    """
    Entry class containing the player object and its current status at the table.
    """
    player: Player
    status: Status


class UserList(BaseModel):
    """
    User handler for tables.
    """
    users: Dict[str, UserListEntry] = {}

    def __len__(self):
        """Amount of players"""
        return len(self.users)

    @property
    def players(self) -> list[Player]:
        """
        Returns all players at the table.
        :return: players of the table
        :rtype: list[Player]
        """
        users = [user.player for user in self.users.values()]
        return users

    @property
    def ready(self) -> bool:
        """
        Returns if all players are ready.
        :return: Ready or not
        :rtype: boolean
        """
        for player in self.users.values():
            if not player.status.ready:
                return False
        return True

    @property
    def teams(self) -> list[Team]:
        """
        Returns the teams.
        :return: list of teams
        """
        players_by_team = list(sorted(self.users.values(), key=lambda x: x.status.team))
        player_by_team: list[list[Player]] = [[entry.player for entry in list(grp)]
                                              for k, grp in groupby(players_by_team,
                                                                    lambda x: x.status.team)]
        teams: list[Team] = [Team(players=players) for players in player_by_team]
        return teams

    def team(self, player: Player) -> Optional[int]:
        """
        Gets the id of the team for a player.
        :param player: for which the id should be retrieved
        :type player: Player
        :return: Integer if player joined a team or None if not.
        :rtype: int
        """
        status: Status = self._get_status(player)
        return status.team

    def team_size(self, team: int) -> int:
        """
        Gets the size of the team.
        :param team: ID of the team
        :type team: int
        :return: Amount of members
        :rtype: int
        """
        return len([entry for entry in self.users.values() if entry.status.team == team])

    def is_joined(self, player: Player) -> bool:
        """
        Checks if the player is already at the table.
        :param player: to check
        :type player: Player
        :return: True if is member else false
        :rtype: bool
        """
        return player.username in self.users

    def append(self, player: Player):
        """
        Adds a player to the list.
        :param player: player to join
        :type player: Player
        :return: None
        :rtype: None
        """
        if not self.is_joined(player):
            self.users.update({player.username: UserListEntry(player=player, status=Status())})

    def remove(self, player: Player):
        """
        Removes the player from the list.
        :param player: player to leave
        :type player: Player
        :return: None
        :rtype: None
        """
        if self.is_joined(player):
            self.users.pop(player.username)

    def apply_distribution(self, distribution: Distribution) -> None:
        """
        Apply the changes of teams.
        :param distribution: matrix of player assignment to teams
        :return: None
        """
        for entry in distribution:
            self.change_team(self.players[entry.player_index], entry.team_id)

    def change_team(self, player: Player, team: int) -> None:
        """
        Player changes teams.
        :param player: to change teams
        :type player: Player
        :param team: id of the new team
        :type team: int
        :return: None
        :rtype: None
        """
        status: Status = self._get_status(player)
        status.team = team

    def player_ready(self, player: Player):
        """
        Player says they is ready.
        :param player: player who is ready, must be joined
        :type player: Player
        :return: Raised PlayerNotJoinedError if the player has not yet joined.
        :rtype: None
        """
        if not self.is_joined(player):
            raise PlayerNotJoinedError()
        status: Status = self._get_status(player)
        status.ready = True

    def player_unready(self, player: Player):
        """
        Player says they is not ready.
        :param player: player who is not ready
        :type player: Player
        :return: Raised PlayerNotJoinedError if the player has not yet joined.
        :rtype: None
        """
        if not self.is_joined(player):
            raise PlayerNotJoinedError()
        status: Status = self._get_status(player)
        status.ready = False

    def _get_status(self, player) -> Status:
        return self._get_entry(player).status

    def _get_entry(self, player) -> UserListEntry:
        return self.users[player.username]



"""Checks the legality of a move."""
from typing import Optional

from whist_core.cards.card import Card
from whist_core.cards.card_container import UnorderedCardContainer


# pylint: disable=too-few-public-methods
class LegalChecker:
    """
    Static legal checker.
    """

    @staticmethod
    def check_legal(hand: UnorderedCardContainer, card: Card, first: Optional[Card]) -> bool:
        """
        Checks if move is legal.
        :param hand: of the current player
        :param card: the card which should be played next
        :param first: the first played card, can be None if no card has been played
        :return: True if legal else false
        """
        first_card_played = first is not None
        if not first_card_played:
            return True
        if card.suit == first.suit:
            return True
        return not hand.contains_suit(first.suit)



"""Player instance during the game phase."""
from pydantic import BaseModel

from whist_core.cards.card_container import UnorderedCardContainer
from whist_core.user.player import Player


class PlayerAtTable(BaseModel):
    """
    Wraps the current hand and player instance.
    """
    player: Player
    hand: UnorderedCardContainer
    team: int

    def __eq__(self, other):
        """Checks if the other is equal."""
        if not isinstance(other, PlayerAtTable):
            return False
        return self.player == other.player

    def __repr__(self):
        """String representation"""
        return f'PlayerAtTable: {self.player}'



"""Card related classes"""

from enum import Enum
from functools import total_ordering
from typing import Any, Optional, Iterator

import deprecation
from pydantic import BaseModel

from whist_core.util import enforce_str_on_dict


@total_ordering
class _CardEnum(Enum):
    def __new__(cls, *args) -> '_CardEnum':
        obj = object.__new__(cls)
        obj._value_ = args[0]
        # pylint: disable=protected-access, unused-private-member
        obj.__ordinal = len(cls.__members__)
        return obj

    def __init__(self, _: str, short_name: Optional[str] = None):
        self.__short_name = short_name

    @classmethod
    def _missing_(cls, value: Any) -> Optional['_CardEnum']:
        if isinstance(value, int):
            try:
                return list(cls.__members__.values())[value]
            except IndexError:
                pass
        elif isinstance(value, str):
            for member in cls.__members__.values():
                if value in (member.long_name, member.short_name):
                    return member

        return None

    @property
    def ordinal(self) -> int:
        """
        Get the ordinal.

        :return: ordinal
        """
        return self.__ordinal

    @property
    def long_name(self) -> str:
        """
        Get the long version of the name.

        :return: long version of name
        """
        return self.value

    @property
    def short_name(self) -> str:
        """
        Get the short version of the name.

        :return: short version of name
        """
        return self.__short_name

    def __lt__(self, other: Any) -> bool:
        if self.__class__ is other.__class__:
            return self.ordinal < other.ordinal
        return NotImplemented

    def __str__(self) -> str:
        return self.value


class Suit(_CardEnum):
    """Suits in a playing card deck"""

    CLUBS = ('clubs', '♣')
    DIAMONDS = ('diamonds', '♦')
    HEARTS = ('hearts', '♥')
    SPADES = ('spades', '♠')


class Rank(_CardEnum):
    """Ranks in a playing card deck"""

    NUM_2 = '2'
    NUM_3 = '3'
    NUM_4 = '4'
    NUM_5 = '5'
    NUM_6 = '6'
    NUM_7 = '7'
    NUM_8 = '8'
    NUM_9 = '9'
    NUM_10 = '10'
    J = ('jack', 'J')
    Q = ('queen', 'Q')
    K = ('king', 'K')
    A = ('ace', 'A')


@total_ordering
class Card(BaseModel, frozen=True):
    """A playing card"""

    suit: Suit
    rank: Rank

    @staticmethod
    def all_cards() -> Iterator['Card']:
        """
        Get iterator of all cards.

        :return: all cards
        """
        return (Card(suit=suit, rank=rank) for suit in Suit for rank in Rank)

    @property
    def short_name(self) -> str:
        """
        Get the short name of this card.

        :return: short name
        """
        short_name = self.rank.short_name \
            if self.rank.short_name is not None \
            else self.rank.long_name
        return f'{self.suit.short_name}{short_name}'

    @property
    def name(self) -> str:
        """
        Get the name of this card.

        :return: name
        """
        return f'{self.rank} of {self.suit}'

    @deprecation.deprecated("Use model_dump instead. Will be removed in V1.")
    def dict(self, *args, **kwargs):
        """
        Returns the dictionary. See BaseModel for details.
        """
        super_dict = super().model_dump(*args, **kwargs)
        return enforce_str_on_dict(super_dict, ('suit', 'rank'))

    def model_dump(self, *args, **kwargs):
        """
        Returns the dictionary. See BaseModel for details.
        """
        super_dict = super().model_dump(*args, **kwargs)
        return enforce_str_on_dict(super_dict, ('suit', 'rank'))

    def __lt__(self, other: Any) -> bool:
        """Checks if the other card is lower than this card."""
        if self.__class__ is other.__class__:
            return (self.suit, self.rank) < (other.suit, other.rank)
        return NotImplemented

    def __str__(self) -> str:
        """Returns string representation of this card."""
        return self.name



"""DAO of session."""
from typing import Any, Union

from pydantic import model_validator
from typing_extensions import Literal

from whist_core.error.table_error import TableFullError, TeamFullError, TableNotReadyError, \
    TableNotStartedError, TableSettingsError
from whist_core.game.errors import RubberNotDoneError
from whist_core.game.rubber import Rubber
from whist_core.session.matcher import Matcher, subclass_registry
from whist_core.session.session import Session
from whist_core.user.player import Player


class Table(Session):
    """
    The game logics instance of a room to play Whist.
    """
    min_player: int
    max_player: int
    team_size: int = 2
    started: bool = False
    rubbers: list[Rubber] = []
    matcher: Matcher

    # copied from
    # https://blog.devgenius.io/deserialize-child-classes-with-pydantic-that-gonna-work-784230e1cf83
    def __init__(self, **data: Any) -> None:
        """
        Constructor
        :param data: fields from above
        """
        matcher = data['matcher']
        if isinstance(matcher, dict):
            item_matcher_keys = sorted(matcher.keys())
            for _, subclass in subclass_registry.items():
                matcher_keys = sorted(subclass.__fields__.keys())
                if item_matcher_keys == matcher_keys:
                    matcher = subclass(**matcher)
                    break
            data['matcher'] = matcher
        super().__init__(**data)

    def model_dump(self, *, mode: Union[Literal['json', 'python'], str] = 'python', include=None,
                   exclude=None, by_alias: bool = False, exclude_unset: bool = False,
                   exclude_defaults: bool = False, exclude_none: bool = False,
                   round_trip: bool = False, warnings: bool = True) -> dict[str, Any]:
        """
        Overrides model_dump to ensure matcher is correctly dumped.
        :param mode:
        :param include:
        :param exclude:
        :param by_alias:
        :param exclude_unset:
        :param exclude_defaults:
        :param exclude_none:
        :param round_trip:
        :param warnings:
        :return:
        """
        model = super().model_dump(mode=mode, include=include, exclude=exclude, by_alias=by_alias,
                                   exclude_unset=exclude_unset, exclude_defaults=exclude_defaults,
                                   exclude_none=exclude_none, round_trip=round_trip,
                                   warnings=warnings)
        model['matcher'] = self.matcher.model_dump()
        return model

    # pylint: disable=no-self-argument
    @model_validator(mode="before")
    @classmethod
    def validate_min_is_lower_max_player(cls, values):
        """
        Checks if the min_player is less or equal than max_player.
        :param values:
        :return:
        """
        if values.get('min_player') > values.get('max_player'):
            raise TableSettingsError('The amount of minimum player must not be higher than the '
                                     'maximum amount.')
        return values

    def __len__(self):
        """
        The amount of players joined.
        :return: # player
        :rtype: int
        """
        return len(self.users)

    @property
    def ready(self) -> bool:
        """
        Flag if the table is ready to start playing.
        :return Ready or not
        :rtype: boolean
        """
        return len(self.users) >= self.min_player and self.users.ready

    @property
    def current_rubber(self) -> Rubber:
        """
        Returns the current rubber
        :return: the latest rubber entry
        """
        if not self.started:
            raise TableNotStartedError()
        return self.rubbers[-1]

    def next_rubber(self) -> Rubber:
        """
        Creates the next rubber. In order to create the first rubber use 'Table.start()'.
        :return: the new rubber
        """
        if len(self.rubbers) == 0:
            raise TableNotStartedError()
        if self.rubbers[-1].done:
            self.rubbers.append(self._create_rubber())
        else:
            raise RubberNotDoneError()
        return self.current_rubber

    def start(self) -> None:
        """
        Starts the table, but will check if every player is ready first.
        """
        if not self.ready:
            raise TableNotReadyError()

        self.rubbers.append(self._create_rubber())
        self.started = True

    def join(self, player: Player) -> None:
        """
        If a seat is available a player joins the table.
        :param player: who wants to join the table
        :type player: Player
        :return: None or raised an error if the table is already full.
        :rtype: None
        """
        if len(self.users) < self.max_player:
            self.users.append(player)
        else:
            raise TableFullError(f'Table with name: {self.name} is already full.')

    def leave(self, player: Player) -> None:
        """
        Remove a player from table.
        :param player: The player to remove.
        :type player: Player
        :return: None
        :rtype: None
        """
        self.users.remove(player)

    def join_team(self, player: Player, team: int) -> None:
        """
        Player joins a team.
        :param player: to join a team
        :type player: Player
        :param team: id of the new team
        :type team: int
        :return: None if successful or raises Error if team is full
        :rtype: None
        """
        if not self.users.is_joined(player):
            self.join(player)
        team_size = self.users.team_size(team)
        if team_size >= self.team_size:
            raise TeamFullError(f'Team with id: {team} is already full.')
        self.users.change_team(player, team)

    def player_ready(self, player) -> None:
        """
        Player says they is ready.
        :param player: player who is ready
        :type player: Player
        :return: None
        :rtype: None
        """
        self.users.player_ready(player)

    def player_unready(self, player) -> None:
        """
        Player says they is not ready.
        :param player: player who is not ready
        :type player: Player
        :return: None
        :rtype: None
        """
        self.users.player_unready(player)

    def _create_rubber(self):
        distribution = self.matcher.distribute(users=self.users)
        self.users.apply_distribution(distribution)
        teams = self.users.teams
        rubber = Rubber(teams=teams)
        return rubber



"""DAO of team."""
from pydantic import BaseModel

from whist_core.user.player import Player


class Team(BaseModel):
    """
    Data wrapper for a team.
    """
    players: list[Player]

    def __hash__(self):
        """Hash value of the team object."""
        return hash(tuple(self.players))

    @property
    def rating(self) -> float:
        """
        The average Elo-Rating of the team members.
        :return: average team rating
        :rtype: float
        """
        return sum(player.rating for player in self.players) / len(self.players)

    def games_played(self, amount: int = 1) -> None:
        """
        Increases the games played counter for each team member.
        :param amount: The amount of new games played. Default 1
        :type amount: int
        :return: None
        :rtype: None
        """
        for player in self.players:
            player.games += amount



"""
Result after one hand.
"""
from typing import Any

from pydantic import BaseModel

from whist_core.scoring.team import Team


class Score(BaseModel):
    """
    Score of a hand being played.
    """
    hand_score: dict = {}

    def __init__(self, teams: list[Team], scores: list[int], **data: Any):
        """
        Constructor.
        :param teams: list of the teams
        :param scores: list if the scores as int
        """
        super().__init__(**data)
        for team, score in zip(teams, scores):
            self.hand_score.update({team: score})
            team.games_played()

    def __getitem__(self, item):
        """Returns a specific hand score."""
        return self.hand_score[item]

    @property
    def winner(self) -> Team:
        """
        Returns the winner team.
        :rtype: Team
        """
        return max(self.hand_score.items(), key=lambda x: x[1])[0]

    def won(self, team: Team) -> bool:
        """
        Check if the team won that round.
        :param team: Team for which to check if has won.
        :type team: Team
        :return: True if won, else false
        :rtype: bool
        """
        return self.hand_score[team] == max(self.hand_score.values())



"""Scoring util"""
from whist_core.game.hand import Hand
from whist_core.game.play_order import PlayOrder
from whist_core.scoring.score import Score
from whist_core.scoring.team import Team

# A team score for each trick exceeding 6.
TRICK_EXCESS_BASE = 6


class ScoreCalculator:
    """
    Util class for score calculating.
    """

    @staticmethod
    def calc_score(hand: Hand, play_order: PlayOrder) -> Score:
        """
        Calculates the score at the end of the hand.
        :param hand: the hand to be scored
        :param play_order: the order of players
        :return: Score
        """
        tricks_won = ScoreCalculator.calc_score_raw(hand)
        players_by_team = play_order.to_team_list()
        teams = [Team(players=players) for players in players_by_team]
        score = Score(teams=teams, scores=tricks_won)
        return score

    @staticmethod
    def calc_score_raw(hand: Hand) -> list[int]:
        """
        Calculates the raw score values by team.
        :param hand: the hand to be scored
        :return: list of integers representing the score
        """
        tricks_won = [0, 0]
        for trick in hand.tricks:
            winner = trick.winner
            tricks_won[winner.team] += 1
        return [max(0, trick_score - TRICK_EXCESS_BASE) for trick_score in tricks_won]



"""
Retrieves envs for the whist core module.
"""
import os

# remember to also update the version in pyproject.toml!
__version__ = '0.9.1'

ALGORITHM = os.getenv('ALGORITHM', 'HS256')
SECRET_KEY = os.getenv('SECRET_KEY', 'geheim')



"""Warnings during game phase."""


class TrickNotDoneWarning(Warning):
    """The current trick is not done yet."""


class ServSuitFirstWarning(Warning):
    """Player must serv the suit of the lead."""



"""
Scores over several hands.
"""
from pydantic import BaseModel

from whist_core.scoring.score import Score
from whist_core.scoring.team import Team


class ScoreCard(BaseModel):
    """
    Collects the results of several hands.
    """
    hands: list[Score] = []

    def __len__(self):
        """Amount of hands played."""
        return len(self.hands)

    @property
    def max(self) -> int:
        """
        Returns the highest amount of hands won by either team.
        :rtype: int
        """
        score_by_team: dict[Team, int] = {}
        for hand in self.hands:
            team = hand.winner
            if team in score_by_team:
                score_by_team[team] += 1
            else:
                score_by_team[team] = 1
        return 0 if len(score_by_team) == 0 else max(score_by_team.values())

    def add_score(self, score: Score) -> None:
        """
        Add the score of one hand.
        :param score: Score after one hand played
        :type score: Score
        :return: None
        :rtype: None
        """
        self.hands.append(score)

    def score(self, team: Team) -> int:
        """
        Getter for how many hands have been won by a team.
        :param team: for whom to look
        :type team: Team
        :return: Amount of hands won.
        :rtype: int
        """
        return sum(hand.hand_score[team] for hand in self.hands if hand.won(team))

    def won(self, team) -> int:
        """
        Check if the team won more hands.
        :param team: Team for which to check.
        :type team: Team
        :return: 1 if the team won more hands. 0 if they lost as many as they won. -1 if the lost
        more games than won.
        :rtype: int
        """
        score = self.score(team)
        other_team = self._other_team(team)
        other_score = self.score(other_team)
        if score > other_score:
            return 1
        if score == other_score:
            return 0
        return -1

    def _other_team(self, team: Team) -> Team:
        teams = list(self.hands[0].hand_score.keys())
        teams.remove(team)
        other_team = teams[0]
        return other_team



"""Hand of whist"""
from typing import Optional, Any, Dict

import deprecation
from pydantic import BaseModel

from whist_core.cards.card import Card, Suit
from whist_core.cards.card_container import UnorderedCardContainer, OrderedCardContainer
from whist_core.game.errors import HandDoneError
from whist_core.game.play_order import PlayOrder
from whist_core.game.player_at_table import PlayerAtTable
from whist_core.game.trick import Trick
from whist_core.game.warnings import TrickNotDoneWarning
from whist_core.util import enforce_str_on_dict


class Hand(BaseModel):
    """
    Hand of whist.
    """
    tricks: list[Trick] = []
    trump: Suit

    def done(self) -> bool:
        """
        Check if the hand is done.
        :return: True if the hand is done, else False
        :rtype: bool
        """
        return len(self.tricks) == 13 and self.tricks[-1]

    @property
    def current_trick(self):
        """
        Returns the current trick.
        """
        return self.tricks[-1]

    @staticmethod
    def deal(play_order: PlayOrder) -> 'Hand':
        """
        Deals the hand and starts the first trick.
        :return: the first trick
        :rtype: Trick
        """
        deck = UnorderedCardContainer.full()
        card: Optional[Card] = None
        while deck:
            player = play_order.get_next_player()
            card = deck.pop_random()
            player.hand.add(card)
        trump = card.suit

        first_trick = Trick(play_order=list(play_order), stack=OrderedCardContainer.empty(),
                            trump=trump)
        hand = Hand(tricks=[first_trick], trump=trump)
        return hand

    def next_trick(self, play_order: PlayOrder) -> Trick:
        """
        Starts the next trick.
        :return: the next trick
        :rtype: Trick
        """
        if self.done():
            raise HandDoneError()
        if len(self.tricks) == 0:
            next_trick_order = play_order
        elif self.tricks[-1].done:
            next_trick_order = self._winner_plays_first_card(play_order)
        else:
            raise TrickNotDoneWarning()
        next_trick = Trick(play_order=list(next_trick_order), trump=self.trump)
        self.tricks.append(next_trick)
        return next_trick

    @deprecation.deprecated("Use 'model_dump()' instead")
    def dict(self, *args, **kwargs):
        """Returns as dictionary."""
        super_dict = super().dict(*args, **kwargs)
        return enforce_str_on_dict(super_dict, ['trump'])

    def model_dump(self, *, mode: str = 'python', include=None,
                   exclude=None, by_alias: bool = False, exclude_unset: bool = False,
                   exclude_defaults: bool = False, exclude_none: bool = False,
                   round_trip: bool = False, warnings: bool = True) -> Dict[str, Any]:
        """Returns as dictionary."""
        model = super().model_dump(mode=mode, include=include, exclude=exclude, by_alias=by_alias,
                                   exclude_unset=exclude_unset, exclude_defaults=exclude_defaults,
                                   exclude_none=exclude_none, round_trip=round_trip,
                                   warnings=warnings)
        return enforce_str_on_dict(model, ['trump'])

    def _winner_plays_first_card(self, play_order: PlayOrder) -> PlayOrder:
        winner: PlayerAtTable = self.tricks[-1].winner
        return play_order.rotate(winner)



"""DTO of a game room"""
from pydantic import BaseModel

from whist_core.session.userlist import UserList


class Session(BaseModel):
    """
    User can join to play a game of Whist.
    """
    name: str
    users: UserList = UserList()



"""DAO of user."""
from typing import Optional

from pydantic import field_validator

from whist_core.error.player_error import NegativeRatingError
from whist_core.user.user import User


class Player(User):
    """
    This is the server side class of an user.
    """
    games: int = 0
    rating: int

    def __str__(self):
        """Returns string representation of the player."""
        return self.username

    def __hash__(self):
        """Returns the hash value of the player object."""
        return hash(self.username)

    def __eq__(self, other):
        """Checks if the other is the same player."""
        if not isinstance(other, Player):
            return False
        return other.username == self.username

    # Pydantic will convert this into a classmethod, cls is the correct parameter
    @field_validator('rating')
    @classmethod
    # noinspection ImproperFirstParameter
    def rating_must_not_be_negative(cls, value):  # pylint: disable=no-self-argument
        """
        Validates the rating. It must be zero or positive.
        """
        if value < 0:
            raise NegativeRatingError()
        return value

    @staticmethod
    def get_player(database: dict, username: str) -> Optional['Player']:
        """
        Returns a player for a given username if they are in the given database.
        :param database: where to look for the user
        :type database: dictionary
        :param username: the name of the user to look for
        :type username: string
        :return: The player instance or None
        :rtype: Player or None
        """
        if username in database:
            return database[username]
        return None



"""One Game of whist"""

from pydantic import BaseModel

from whist_core.game.errors import HandNotDoneError
from whist_core.game.hand import Hand
from whist_core.game.play_order import PlayOrder
from whist_core.game.player_at_table import PlayerAtTable
from whist_core.scoring.score_calculator import ScoreCalculator
from whist_core.scoring.score_card import ScoreCard
from whist_core.user.player import Player


class Game(BaseModel, arbitrary_types_allowed=True):
    """
    One Game of whist.
    """

    play_order: PlayOrder
    win_score: int = 3
    score_card: ScoreCard = ScoreCard()
    hands: list[Hand] = []

    def next_hand(self) -> Hand:
        """
        Checks if the current hand is done and if so will return the next hand. If not it will
        return the current hand.
        :rtype: Hand
        """
        if self.current_hand is None:
            hand = Hand.deal(self.play_order)
            self.hands.append(hand)
        elif not self.current_hand.done():
            raise HandNotDoneError()
        else:
            score = ScoreCalculator.calc_score(self.current_hand, self.play_order)
            self.score_card.add_score(score)
            self._next_play_order()
            hand = Hand.deal(self.play_order)
            self.hands.append(hand)
        return self.current_hand

    @property
    def current_hand(self):
        """
        Returns the current hand if there is one. Else None.
        """
        return self.hands[-1] if len(self.hands) > 0 else None

    @property
    def done(self):
        """
        Check if game is done.
        :return: True if done else false
        :rtype: bool
        """
        return self.win_score <= self.score_card.max

    def get_player(self, player: Player) -> PlayerAtTable:
        """
        Retrieves the PlayerAtTable for the player given.
        :param player: who needs it's counterpart at the table
        :return: the player at table
        """
        return self.play_order.get_player(player)

    def _next_play_order(self) -> None:
        """
        Creates the next order of player for next hand.
        """
        self.play_order = self.play_order.next_order()
